#include <iostream>
#include <vector>
#include <set>

using namespace std;

class Item {
public:
    enum ItemType {
        INT32,
        INT64,
        STRING
    };
    
    ItemType type;
    string key;
    string description;

    
    Item( ItemType pType, string pKey, string pDescription ) {
        type = pType;
        key = pKey;
        description = pDescription;
    }
};

class OutputFormat {
private:
    static int nextFormatID;
public:
    
    int formatID;
    
    string formatKey;
    string formatDescription;
    vector<Item> vectorItems;
    
    OutputFormat( string pFormatKey, string pFormatDescription ) {
        cout << "Constructeur OutputFormat" << endl;
        
        formatID = nextFormatID++;
        
        formatKey = pFormatKey;
        formatDescription = pFormatDescription;
    }
    ~OutputFormat() {
        cout << "Destructeur OutputFormat" << endl;
    }
    
    void addItem( Item::ItemType pType, string pKey, string pDescription ) {
        vectorItems.push_back( Item(pType, pKey, pDescription) );
    }
    
    void print() {
        cout << formatKey << " : " << formatDescription << endl;
        for ( vector<Item>::iterator it = vectorItems.begin(); it != vectorItems.end(); it++ ) {
            cout << "  " << it->key << " " << it->description << endl;
        }
    }
};

class Output {
private:
    set<int> knownFormats;
public:
    FILE* outputFile;
    
    Output() {
        cout << "Constructeur Output" << endl;
    }
    
    ~Output() {
        cout << "Destructeur Output" << endl;
    }
    
    
    void create(string fileName) {
        outputFile = fopen( fileName.c_str(), "w" );
    }
    
    void close() {
        fclose( outputFile );
    }
    
    void log( OutputFormat &format...) {
        va_list args;
        va_start(args, format);
        string buffer;
        
        if ( knownFormats.find( format.formatID ) == knownFormats.end() ) {
            knownFormats.insert( format.formatID );
            buffer = "#" + to_string(format.formatID) + " " + format.formatKey + " " + format.formatDescription;
            for ( vector<Item>::iterator it = format.vectorItems.begin(); it != format.vectorItems.end(); it++ ) {
                buffer += " (" + to_string(it->type);
                buffer += "," + it->key + "," + it->description + ")";
            }
            buffer += "\n";
            fprintf( outputFile , buffer.c_str());
        }

        
        buffer = to_string( format.formatID );
        
        for ( vector<Item>::iterator it = format.vectorItems.begin(); it != format.vectorItems.end(); it++ ) {
            switch ( it->type ) {
                case Item::ItemType::INT32:
                    buffer += " " + to_string( va_arg(args, int) );
                    break;
                    
                default:
                    break;
            }
        }
        cout << buffer << endl;
        buffer += "\n";
        fprintf( outputFile , buffer.c_str());
    }
    
};

int OutputFormat::nextFormatID = 0;

int main( int argc, const char* argv[] ) {
    cout << "Debut" << endl;
    
    OutputFormat formatInitSLR( "InitSLR", "Initilisation SLR" );
    formatInitSLR.addItem( Item::INT32, "SLRX", "SLR X" );
    formatInitSLR.addItem( Item::INT32, "SLRY", "SLR Y" );
    formatInitSLR.addItem( Item::INT32, "SLRZ", "SLR Z" );
    formatInitSLR.addItem( Item::INT64, "anID", "anchor node ID" );
    formatInitSLR.addItem( Item::INT64, "rnID", "receiver node ID" );
    formatInitSLR.print();
    
    OutputFormat formatDeDenFinish( "DeDenFinish", "Deden estimation finished" );
    formatDeDenFinish.addItem( Item::INT64, "nID", "node id" );
    formatDeDenFinish.addItem( Item::INT32, "rnum", "last round number" );
    formatDeDenFinish.addItem( Item::INT32, "est", "neighborhood cardinality estimation" );
    
    Output output;
    output.create( "all.log" );
    
    output.log( formatInitSLR, 10, 11, 12, 0, 42 );
    output.log( formatDeDenFinish, 150, 5, 1900);
    output.log( formatInitSLR, 10, 11, 12, 0, 43 );
    output.log( formatDeDenFinish, 160, 5, 1900);
    output.log( formatDeDenFinish, 170, 5, 1900);
    
    output.close();
    
//    //////////////////////////////
//    // *** Ecriture ***
//    //////////////////////////////
//    
//    //
//    // On défini un ou plusieurs formats de ligne de log
//    // Chaque format est une liste ordonnée de propriétés ayant chacune un type (int, float, string, ...).
//    // La liste est initialement vide, il faut la construire avant de l'utiliser.
//    //
//    
//    // OutputFormat( clef, description )
//    // addItem( type, clef, description )
//
//    OutputFormat formatInitSLR( "InitSLR", "Initilisation SLR" );
//    formatInitSLR.addItem( OutputFormat::dataFormat::int32, "SLRX", "SLR X" );
//    formatInitSLR.addItem( "int", "SLRX", "SLR X" );
//    formatInitSLR.addItem( "int", "SLRZ", "SLR Z" );
//    formatInitSLR.addItem( "int", "anID", "anchor node ID" );
//    formatInitSLR.addItem( "int", "rnID", "receiver node ID" );
//    
//    OutputFormat formatDeDenFinish( "DeDenFinish", "Deden estimation finished" );
//    formatDeDenFinish.addItem( "int", "nID", "node id" );
//    formatDeDenFinish.addItem( "int", "rnum", "last round number" );
//    formatDeDenFinish.addItem( "int", "est", "neighborhood cardinality estimation" );
//    
//    
//    [...]
//    
//    // On déclare et on ouvre un fichier de sortie (peut aussi être la sortie standard)
//    
//    Output output();
//    output.create( "all.log" );
//    
//    [...]
//    
//    //
//    // On écrit dans le fichier de sortie grâce à une fonction utilisant une function variadic (nombre de paramètre variable)
//    // Le premier paramètre est une format de ligne précédemment déclaré
//    // Les paramètres suivants sont les valeurs à sauver (en respectant l'ordre de la déclaration, comme un printf)
//    // Des vérifications de type et de nombre de paramètres peuvent facilement être faites
//    //
//    
//    output.log( formatInitSLR, node.SLRX, node.SLRY, node.SLRZ, beacon.anchorID, node.nodeId );
//    
//    //
//    // Rien n'empèche de sauver dans un même fichier des lignes de formats différents.
//    //
//    
//    output.log( formatDeDenFinish, node.nodeId, lastRound, estimation );
//    
//    output.close();
//    
//    
//    // Le fichier de sortie doit contenir un index des formats de ligne utilisés.
//    // Cet index peut être mis en début de fichier (mais pour simplifier le travail, il faut alors probablement réserver un espace fixe en début de fichier).
//    
//    //////////////////////////////
//    // *** Lecture ***
//    //////////////////////////////
//    
//    Output output();
//    output.open( "all.log" );
//
//    LogLine *ligne = new LogLine();
//    
//    while ( output.readLine( ligne ) != NULL ) {
//        if ( ligne.formatDigest() == "InitSLR" ) {
//            cout << "SLR_X:" << ligne.getItem("SLRX") << endl;
//            cout << "SLR_Y:" << ligne.getItem("SLRY") << endl;
//            cout << "SLR_Z:" << ligne.getItem("SLRZ") << endl;
//        }
//        
//        // ou
//        
//        if ( ligne.formatDigest == "InitSLR" ) {
//            int slrX, slrY, slrZ;
//            int anchorId, receiverId;
//            ligne.parse( &slrX, &slrY, &slrZ, &anchorId, &receiverId );
//            cout << slrX << "," << slrY << "," << slrZ << endl;
//        }
//        
//        // ou
//        
//        if ( ligne.hasItem("SLRX") ) {
//            [...]
//        }
//        
//    }

}