/*
 * testChevauche.cpp
 *
 *  Created on: 11 juil. 2017
 *      Author: ddhoutau
 */

#include <math.h>
#include <iostream>
#include <fstream>
#include <gmpxx.h>
#include <random>
#include <vector>
#include <map>

#include <tclap/CmdLine.h>

using namespace std;

int main(int _argc, char **_argv) {

    long int nbNodes;
    long int windowWidth, packetSize;
    long int nbTirages;
    long int nbRetrans;
    long int redondance;
    long int RNGSeed;
    vector<long int> vectRetransDistribution(1000,0);


    try {
        TCLAP::CmdLine cmd("testChevauche", ' ', "1.1");
        TCLAP::ValueArg<long int> nbNodesParam("n", "nbNodes", "nombre de noeuds", false, 3, "long int");
        TCLAP::ValueArg<long int> windowWidthParam("w", "windowWidth", "largeur de la fenêtre", false, 100, "long int");
        TCLAP::ValueArg<long int> packetSizeParam("p", "packetSize", "durée du paquet (temps de vol inclu)", false, 10, "long int");
        TCLAP::ValueArg<long int> nbTiragesParam("", "nbTirages", "nombre de tirages", false, 1000, "long int");
        TCLAP::ValueArg<long int> redondanceParam("", "redondance", "redondance désirée", false, 1, "long int");
        TCLAP::ValueArg<long int> RNGSeedParam("", "RNGSeed", "desired RNG seed", false, 0, "long int");

        cmd.add(nbNodesParam);
        cmd.add(windowWidthParam);
        cmd.add(packetSizeParam);
        cmd.add(nbTiragesParam);
        cmd.add(redondanceParam);
        cmd.add(RNGSeedParam);

        cmd.parse( _argc, _argv );

        nbNodes = nbNodesParam.getValue();
        windowWidth = windowWidthParam.getValue();
        packetSize = packetSizeParam.getValue();
        nbTirages = nbTiragesParam.getValue();
        redondance = redondanceParam.getValue();
        RNGSeed = RNGSeedParam.getValue();

    } catch (TCLAP::ArgException &e) {
        cerr << "*** ERROR *** " << e.error() << " for arg " << e.argId() << std::endl;
    }

    std::mt19937 gen(RNGSeed);
    std::uniform_int_distribution<long int> dis(0, windowWidth-packetSize);

    cout << "# nbNodes:     " << nbNodes << endl;
    cout << "# windowWidth: " << windowWidth << endl;
    cout << "# packetSize:  " << packetSize << endl;
    cout << "# redondance:  " << redondance << endl;
    cout << "# nbTirages:   " << nbTirages << endl;

    vector<long int>datesDebut;
    long int nbConcurrent;
    bool fin;
    long int currentDate;
    long int n;
    multimap<long int, char> eventsMap;

    long int datePremier = 0;
    long int dateRedondance = 0;

    for (long int tirage=0; tirage<nbTirages; tirage++) {
        datesDebut.clear();
        eventsMap.clear();

        for (n=0; n<nbNodes; n++) {
            long int date = dis(gen);
            //cout << "tirage:" << tirage << " packet:" << n << "  debut:" << date << endl;
            datesDebut.push_back(date);
            eventsMap.insert(pair<long int, char>(date,'d'));
            eventsMap.insert(pair<long int, char>(date+packetSize,'f'));
        }
        sort(datesDebut.begin(), datesDebut.end());

        datePremier += eventsMap.begin()->first;

        nbConcurrent = 0;
        fin = false;
        currentDate = datesDebut[0];
        n = 1;
        nbRetrans = 0;
        long int nbFinished = 0;

//        for (auto it=eventsMap.begin(); it != eventsMap.end(); it++) {
//            cout << it->first << " " << it->second << endl;
//        }

        multimap<long int, char>::iterator first;
        while ( !fin ) {
            if (eventsMap.empty() || nbFinished >= redondance) {
                fin = true;
            } else {
                first = eventsMap.begin();
                if (first->second == 'd') {
                    nbRetrans++;
                    if ( nbRetrans == redondance ) {
                        dateRedondance += eventsMap.begin()->first;
                    }
                } else if (first->second == 'f') {
                    nbFinished++;
                }
                eventsMap.erase(first);
            }
        }
//        while (!fin) {
//
////            if (datesDebut[n] < (datesDebut[n-1])+packetSize ) {
////                nbRetrans++;
////            } else {
////                fin = true;
////            }
//
//            if (nbRetrans < redondance) {
//                nbRetrans++;
//            } else {
//                fin = true;
//            }
//
//            n++;
//            if (n >= nbNodes) fin = true;
//        }
        //cout << "tirage:" << tirage << " nbRetrans:" << nbRetrans << endl << endl;
        vectRetransDistribution[nbRetrans]++;
    }

    datePremier /= nbTirages;
    cout << "# datePremier:    " << datePremier << endl;
    dateRedondance /= nbTirages;
    cout << "# dateRedondance: " << dateRedondance << endl;

    cout << "# distribution des retransmissions:" << endl;
    for (int i=0; i<30; i++) {
        cout << i << ": " << vectRetransDistribution[i] << endl;
    }

}



